// vi: set ts=4 sw=4 :
// vim: set tw=75 :

// dllapi.cpp - implementation of Half-Life DLL routines
/*
 * Copyright (c) 2004 OLO
 * This file is a part of PLBot application.
 * You are not allowed to remove these copyrights.
 */

#include <extdll.h>
#include <sdk_util.h>
#include <cbase.h>
#include <player.h>

#include <dllapi.h>
#include <meta_api.h>
#include <sdk_util.h>

#include "bot.h"
#include "bot_nodes.h"
#include "plbot_api.h"  // me

#ifdef MY_DEBUGE_MODE
edict_t *listenserver_edict = NULL;
#endif

DLL_FUNCTIONS other_gFunctionTable;
DLL_GLOBAL const Vector g_vecZero = Vector(0,0,0);

#ifdef MY_DEBUGE_MODE
int m_spriteTexture = 0;
#endif

bool find_objectives = FALSE;
float auto_add = 0.0;

bool bombplanted = FALSE;
bool bombdroped = FALSE;
bool bombnotfound = FALSE;
bool need_to_open_cfg = TRUE;

cm_cvar_t cvar_game[MAX_CVAR];
int cvars_num = 0;
cmd_t cstcmd_names[MAX_CMD];
int cstcmd_num = 0;

int menu_keys[10] = {1,2,4,8,16,32,64,128,256,512};

bool waypoints_loaded;

int DispatchSpawn( edict_t *pent )
{

   if (FClassnameIs(pent,"worldspawn"))
   {
      //precache stuff
#ifdef MY_DEBUGE_MODE
      m_spriteTexture = PRECACHE_MODEL("sprites/dot.spr");//Lasr beam
#endif

      PRECACHE_MODEL("sprites/glow02.spr");//Nodes
      const char* mapname = STRING(gpGlobals->mapname);//Check map type
      if (FStrnEq(mapname, "cs_" , 3))
         Maptype = CS_MAP;//Hostage rescue scenario
      else if (FStrnEq(mapname, "de_" , 3))
         Maptype = DE_MAP;//Demolish scenario map
      else if (FStrnEq(mapname, "as_" , 3))
         Maptype = AS_MAP;//Assasination map
      else if (FStrnEq(mapname, "es_" , 3))
         Maptype = ES_MAP;//Escape scenario map
      else
         Maptype = 0;//Well without proper name of map bots won't play properly

      waypoints_loaded = NodesLoad();//Load nodess
      find_objectives = TRUE;
      cvars_num = 0; //Reset PLBot menu settings
      cstcmd_num = 0;
      number_names = 0;
      chat_num = 0;
      auto_add = 0.0;
   }

   RETURN_META_VALUE(MRES_IGNORED, 0);
}

void DispatchTouch( edict_t *pentTouched, edict_t *pentOther )
{
   if (FStrEq(STRING(pentTouched->v.classname), "player"))
   {
      player_t *pPlayer = GET_EPLAYERPOINTER(pentTouched);

      if (FStrEq(STRING(pentOther->v.classname), "player"))
      {
         player_t *pMate = GET_EPLAYERPOINTER(pentOther);

         if (pPlayer->check_stuck_time < gpGlobals->time)
         {
            pPlayer->action_time = gpGlobals->time + 0.1;
            pPlayer->bot_action = BOT_DUCKING; //Snipe down
            pPlayer->check_stuck_time = gpGlobals->time + 0.3;
            pMate->check_stuck_time = gpGlobals->time + 0.4;
         }
      }
   }

   RETURN_META(MRES_IGNORED);
}

void ClientPutInServer( edict_t *pEntity )
{
   if (waypoints_loaded||!plb_onlywpt->value)
   {
      int player_num = 0;
      for (int a = 1; a <= gpGlobals->maxClients; a++)
      {
         if (IsValidPlayer(GET_IPLAYERPOINTER(a)))
            player_num++;//Count number of players
      }

      //If autoadd on and there is to much players then disconnect one bot
      if (plb_autoadd->value && player_num > plb_autoadd->value)
      {
         player_t *pUsers;
         for (int b = 1; b <= gpGlobals->maxClients; b++)
         {
            pUsers = GET_IPLAYERPOINTER(b);
            if (IsValidPlayer(pUsers) && (pUsers->flags & FL_FAKECLIENT))
            {
               SERVER_COMMAND(UTIL_VarArgs("kick \"%s\"\n",STRING(pUsers->pEdict->v.netname)));
               break;
            }
         }
      }
   }
   RETURN_META(MRES_IGNORED);
}

void ClientDisconnect( edict_t *pEntity )
{
   //Set if can_join so we won't add bots on map change cos then every one is disconnecting
   if (waypoints_loaded||!plb_onlywpt->value)
      auto_add = gpGlobals->time;

   RETURN_META(MRES_IGNORED);
}

BOOL ClientConnect( edict_t *pEntity, const char *pszName, const char *pszAddress, char szRejectReason[ 128 ]  )
{
   player_t *pPlayer = GET_EPLAYERPOINTER(pEntity);
   pPlayer->pEdict = pEntity; //Make for sure that he has entity
   pPlayer->start_action = MSG_CS_TEAM_SELECT;
   pPlayer->curr_weapon = 0;
   pPlayer->can_listen = TRUE;
   pPlayer->botcontrol = 0;
   pPlayer->players_num = 0;
   pPlayer->slap_power = 100;
   pPlayer->flags = 0;
   pPlayer->have_access = FALSE;

#ifdef MY_DEBUGE_MODE

   if (FStrEq(pszAddress,"loopback"))
      listenserver_edict = pEntity; //Only for debug mode on listenserver
#endif

   RETURN_META_VALUE(MRES_IGNORED, TRUE);
}

void ClientUserInfoChanged( edict_t *pEntity, char *infobuffer )
{
   player_t *pPlayer = GET_EPLAYERPOINTER(pEntity);
   pPlayer->pEdict = pEntity; //Make for sure that he has entity

   char *model = INFOKEY_VALUE(infobuffer, "model");
   pPlayer->player_team = CLIENT_UNKNOW;
   if (FStrEq(model, "terror") || FStrEq(model, "arab") || FStrEq(model, "leet")|| FStrEq(model, "arctic") || FStrEq(model, "guerilla"))
      pPlayer->player_team |= CLIENT_TERRORIST;
   else if (FStrEq(model, "urban") || FStrEq(model, "gsg9") || FStrEq(model, "sas")|| FStrEq(model, "gign"))
      pPlayer->player_team |= CLIENT_CT;
   else if (FStrEq(model, "vip"))
      pPlayer->player_team |= (CLIENT_VIP | CLIENT_CT);

   if (plb_password->string && *plb_password->string)
   {
      static char menu_password[256];
      strcpy(menu_password,plb_password->string);
      char *plbot_setinfo = menu_password;
      char *plbot_password = NULL;

      while (plbot_setinfo && *plbot_setinfo)
      {
         if (*plbot_setinfo == ':')
         {
            *plbot_setinfo = '\0';
            plbot_password = plbot_setinfo + 1;
            break;
         }
         plbot_setinfo++;
      }

      if (plbot_password && *plbot_password)
      {
         //If setinfo contains the same as plbot_password then we can be an admin
         if (FStrEq(plbot_password, INFOKEY_VALUE(infobuffer,menu_password)))
         {
            CLIENT_PRINT( pEntity, print_chat, "You have now access to plb_menu\n");
            pPlayer->have_access = TRUE;
         }
         else //Bad password
            pPlayer->have_access = FALSE;
      }
      else //No password so we can be an admin
         pPlayer->have_access = TRUE;

   }
   else
   { //No password so we can be an admin
      pPlayer->have_access = TRUE;
   }


   RETURN_META(MRES_IGNORED);
}

void StartFrame( void )
{
   static player_t *pPlayer = NULL;
   //Search for objectives when all enties loaded
   if (find_objectives)
   {
      find_objectives = FALSE;
      FindObjectives();
      FindRushPointNodes();

      if (plb_config->string && plb_config->string[0])
         SERVER_COMMAND(UTIL_VarArgs("exec %s\n",plb_config->string));

      //Refresh bot time to add
      auto_add = gpGlobals->time;
   }//After plb_jointime fill server with bots
   else if (auto_add && auto_add + plb_jointime->value < gpGlobals->time)
   {
      auto_add = 0.0;
      if (waypoints_loaded||!plb_onlywpt->value)
      {
         int player_num = 0, ct_num = 0, t_num = 0;
         //Count teams and player num
         for (int a = 1; a <= gpGlobals->maxClients; a++)
         {
            pPlayer = GET_IPLAYERPOINTER(a);
            if (IsValidPlayer(pPlayer))
            {
               if (pPlayer->player_team & CLIENT_TERRORIST)
                  t_num++; //Increase number of terrorists
               else
                  ct_num++; //Increase number of ct

               player_num++;
            }
         }
         //If autoadd is set and number of players is the same or
         //is smaler then add CT is t_num is greater or T is not...
         if (plb_autoadd->value && player_num < plb_autoadd->value)
         {
            //We need more bots to add so set next time
            if (plb_autoadd->value - player_num > 1)
               auto_add = gpGlobals->time; //We still need bots to add
            //If there is more T add CT and vice versa
            SERVER_COMMAND((t_num > ct_num)?(char*)"plb_add 2\n":(char*)"plb_add 1\n");
         }
      }
   }

   //Check for planted or droped bomb
   if (Maptype == DE_MAP)
      FindBomb();
   //Check for hostages
   else if (Maptype == CS_MAP)
      CheckHostages();

   if (!plb_pause->value)
   {
      static int bots_number;
      bots_number = 0;
      //Go through bots and "execute their brain"
      for (int a = 1; a <= gpGlobals->maxClients; a++)
      {
         pPlayer = GET_IPLAYERPOINTER(a);
         if (IsValidPlayer(pPlayer) && (pPlayer->flags & FL_FAKECLIENT))
         {
            pPlayer->pEdict->v.flags |= FL_FAKECLIENT;
            BotThink(pPlayer);
            bots_number++;//Increase counter
         }
      }
      if (bots_number != plb_number->value)
         CVAR_DIRSET_STRING(plb_number,UTIL_VarArgs("%d", bots_number));
   }

   RETURN_META(MRES_IGNORED);
}

int developme = 0;

void FindRushPointNodes();

void SlapBot(edict_t *pEntity, edict_t *victim, int power)
{
   if (IsAlive(victim))
   {
      Vector v_forward, v_right;
      ANGLEVECTORS( pEntity->v.angles, v_forward, v_right, NULL );
      victim->v.punchangle.x = 15;
      victim->v.velocity = victim->v.velocity + (v_forward * 220) + Vector(0,0,200);
      EMIT_SOUND_DYN2(victim, CHAN_ITEM, KLAPS_WAVE, 1.0, ATTN_NORM, 0, PITCH_NORM );

      if (victim->v.health <= power)
      {
         victim->v.frags++;
         MDLL_ClientKill(victim);
      }
      else
      {
         victim->v.health -= power;
         victim->v.armorvalue -= power;
         victim->v.dmg_inflictor = victim;
      }
   }
}

void ClientCommand( edict_t *pEntity )
{
   static player_t *pPlayer;
   pPlayer = GET_EPLAYERPOINTER(pEntity);
   static const char *pcmd;
   pcmd = CMD_ARGV(0);
   static const char *arg1;
   arg1 = CMD_ARGV(1);

   // #### STANDARD MENU === BOT MENU ####
   if (FStrEq(pcmd, "menuselect" ))
   {
      int pressed_key = atoi( CMD_ARGV(1) );
      if (pPlayer->start_action == MSG_CS_BOT_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);

         switch (pressed_key)
         {
         case 1:
            BotCreate( pEntity, "1", NULL);
            Botmenu( pPlayer, MSG_CS_BOT_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 2:
            BotCreate( pEntity, "2", NULL);
            Botmenu( pPlayer, MSG_CS_BOT_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 3:
            Botmenu( pPlayer, MSG_CS_SKILL_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 4:
            MenuPlayersList(pPlayer, "Slap Bot Menu", 6, MSG_CS_SLAP_MENU, 0);
            RETURN_META(MRES_SUPERCEDE);
         case 5:
            MenuPlayersList(pPlayer, "Kick Bot Menu", 8, MSG_CS_KICK_MENU, 0);
            RETURN_META(MRES_SUPERCEDE);
         case 6:
            Botmenu( pPlayer, MSG_CS_NODE_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 7:
            MenuSettingsList(pPlayer, 0);
            RETURN_META(MRES_SUPERCEDE);
         case 8:
            MenuCmdList(pPlayer, 0);
            RETURN_META(MRES_SUPERCEDE);
         case 10:
            pPlayer->start_action = MSG_CS_IDLE;
            RETURN_META(MRES_SUPERCEDE);
         }
      }
      //#### SKILL MENU ####
      else if (pPlayer->start_action == MSG_CS_SKILL_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);

         if (pressed_key == 10)
         {
            Botmenu( pPlayer, MSG_CS_BOT_MENU);
            RETURN_META(MRES_SUPERCEDE);
         }

         for (int i=1; i < 33; i++)
         {//Go through 32 palyers
            if (IsValidPlayer(&bots[i])) // is this slot used?
               bots[i].bot_skill = (pressed_key == 6)?RANDOM_LONG(1,5):pressed_key;
         }

         CVAR_SET_FLOAT("plb_skill", pressed_key); //set specified
         Botmenu( pPlayer, MSG_CS_SKILL_MENU);
         RETURN_META(MRES_SUPERCEDE);
      }
      //#### NODES MENU ####
      else if (pPlayer->start_action == MSG_CS_NODE_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);

         switch (pressed_key)
         {
         case 1:
            AddNode(pEntity->v.origin, 0);
            Botmenu( pPlayer, MSG_CS_NODE_MENU);
            CLIENT_PRINT( pEntity, print_center, "Node added\n");
            RETURN_META(MRES_SUPERCEDE);
         case 2:
            if (RemoveNode(pEntity))
               CLIENT_PRINT( pEntity, print_center, "Node removed\n");
            Botmenu( pPlayer, MSG_CS_NODE_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 3:
            if (SetNodesFlag(pEntity, EF_BRIGHTFIELD))
            {
               //Find rush point for all possible kinde of maps
               rushpoint_num = 0; //First reset rushpoint number
               FindRushPointNodes(); //Then find and count
               CLIENT_PRINT( pEntity, print_center, "Rush point set\n");
            }
            Botmenu( pPlayer, MSG_CS_NODE_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 4:
            NodesSave();
            Botmenu( pPlayer, MSG_CS_NODE_MENU);
            CLIENT_PRINT( pEntity, print_center, "Nodes saved\n");
            RETURN_META(MRES_SUPERCEDE);
         case 5:
            if (nodes_on)
            {
               ShowMeNode(FALSE);
               nodes_on = 0;
            }
            else
            {
               ShowMeNode(TRUE);
               nodes_on = 1;
            }

            Botmenu( pPlayer, MSG_CS_NODE_MENU);
            RETURN_META(MRES_SUPERCEDE);
         case 10:
            Botmenu( pPlayer, MSG_CS_BOT_MENU);
            //pPlayer->start_action = MSG_CS_IDLE;
            RETURN_META(MRES_SUPERCEDE);
         }
      }
      //########## GAME SETTINGS
      else  if (pPlayer->start_action == MSG_CS_SET_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);

         int begin_select = 8*pPlayer->menu_state; //Small trick

         if ((pressed_key > 0) && (pressed_key < 9))
         {
            begin_select += pressed_key - 1;

            float cvar_value = (CVAR_GET_FLOAT(cvar_game[begin_select].name));

            if (cvar_value >= cvar_game[begin_select].max)
               cvar_value = 0;
            else
               cvar_value += cvar_game[begin_select].step;

            CVAR_SET_FLOAT(cvar_game[begin_select].name, cvar_value);

            MenuSettingsList(pPlayer, pPlayer->menu_state);
         }
         else if (pressed_key == 9)
         {
            MenuSettingsList(pPlayer, pPlayer->menu_state + 1);
         }
         else if (pressed_key == 10)
         {
            if (pPlayer->menu_state == 0)
               Botmenu( pPlayer, MSG_CS_BOT_MENU);
            else
               MenuSettingsList(pPlayer, pPlayer->menu_state - 1);
         }
         RETURN_META(MRES_SUPERCEDE);
      }
      else if (pPlayer->start_action == MSG_CS_CMD_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);
         int begin_select = 8*pPlayer->menu_state; //Small trick

         if ((pressed_key > 0) && (pressed_key < 9))
         {
            begin_select += pressed_key - 1;
            char*msg= UTIL_VarArgs("%s\n",cstcmd_names[begin_select].cmd);
            MenuCmdList(pPlayer, pPlayer->menu_state);

            char *ptr = NULL;
            while (ptr = strstr(msg, "'"))
               *(ptr) = '\"';

            SERVER_COMMAND(msg); //Execute vote result (command)
         }
         else if (pressed_key == 9)
         {
            MenuCmdList(pPlayer, pPlayer->menu_state + 1);
         }
         else if (pressed_key == 10)
         {
            if (pPlayer->menu_state == 0)
               Botmenu( pPlayer, MSG_CS_BOT_MENU);
            else
               MenuCmdList(pPlayer, pPlayer->menu_state - 1);
         }
         RETURN_META(MRES_SUPERCEDE);
      }
      else if (pPlayer->start_action == MSG_CS_SLAP_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);
         int begin_select = 6*pPlayer->menu_state; //Small trick

         if ((pressed_key > 0) && (pressed_key < 7))
         {
            begin_select += pressed_key - 1;

            SlapBot(pEntity,pPlayer->pList[begin_select], pPlayer->slap_power);
            MenuPlayersList(pPlayer, "Slap Bot Menu", 6, MSG_CS_SLAP_MENU, pPlayer->menu_state);
         }
         else if (pressed_key == 7)
         {
            for (int a = 0; a < pPlayer->players_num;a++)
            {
               if (IsAlive(pPlayer->pList[a]))
                  SlapBot(pEntity,pPlayer->pList[a], pPlayer->slap_power);
            }

            MenuPlayersList(pPlayer, "Slap Bot Menu", 6, MSG_CS_SLAP_MENU, pPlayer->menu_state);
         }
         else if (pressed_key == 8)
         {
            if (pPlayer->slap_power == 0)
               pPlayer->slap_power = 5;
            else if (pPlayer->slap_power == 5)
               pPlayer->slap_power = 20;
            else if (pPlayer->slap_power == 20)
               pPlayer->slap_power = 100;
            else
               pPlayer->slap_power = 0;

            MenuPlayersList(pPlayer, "Slap Bot Menu", 6, MSG_CS_SLAP_MENU, pPlayer->menu_state);
         }
         else if (pressed_key == 9)
         {
            MenuPlayersList(pPlayer, "Slap Bot Menu", 6, MSG_CS_SLAP_MENU, pPlayer->menu_state + 1);
         }
         else if (pressed_key == 10)
         {
            if (pPlayer->menu_state == 0)
               Botmenu( pPlayer, MSG_CS_BOT_MENU);
            else
               MenuPlayersList(pPlayer, "Slap Bot Menu", 6, MSG_CS_SLAP_MENU, pPlayer->menu_state - 1);
         }

         RETURN_META(MRES_SUPERCEDE);
      }
      //######### KICK - SELECT PLAYER
      else if (pPlayer->start_action == MSG_CS_KICK_MENU)
      {
         //Don't allow to manipulate menu by commands
         if (!(pPlayer->menu_keys & (1<<pressed_key)))
            RETURN_META(MRES_SUPERCEDE);
         int begin_select = 8*pPlayer->menu_state; //Small trick
         if ((pressed_key > 0) && (pressed_key < 9))
         {
            begin_select += pressed_key - 1;
            char*msg= UTIL_VarArgs("kick \"%s\"\n", STRING(pPlayer->pList[begin_select]->v.netname));
            SERVER_COMMAND(msg);  // kick the bot using (kick "name")
            Botmenu( pPlayer, MSG_CS_BOT_MENU);
         }
         else if (pressed_key == 9)
         {
            MenuPlayersList(pPlayer, "Kick Bot Menu", 6, MSG_CS_KICK_MENU, pPlayer->menu_state + 1);
         }
         else if (pressed_key == 10)
         {
            if (pPlayer->menu_state == 0)
               Botmenu( pPlayer, MSG_CS_BOT_MENU);
            else
               MenuPlayersList(pPlayer, "Kick Bot Menu", 6, MSG_CS_KICK_MENU, pPlayer->menu_state - 1);
         }
         RETURN_META(MRES_SUPERCEDE);
      }
      else
      {
         pPlayer->start_action = MSG_CS_IDLE;
      }
   }
   else if (FStrEq(pcmd, "plb_menu" ))
   {
      if (pPlayer->have_access)
      {
         if (pPlayer->start_action==MSG_CS_IDLE||pPlayer->start_action==MSG_CS_TEAM_SELECT)
         {
            Botmenu( pPlayer, MSG_CS_BOT_MENU);
         }
         else if (pPlayer->start_action==MSG_CS_CT_SELECT||pPlayer->start_action==MSG_CS_T_SELECT)
         {
            // ############# DONT USE WHEN CLASS SELECTING
         }
         else
         {
            UTIL_ShowMenu(pPlayer->pEdict, 0, 0,FALSE, " " );
            pPlayer->start_action = MSG_CS_IDLE;//Show blank menu
         }
      }
      else
      {
         CLIENT_PRINT( pEntity, print_center, "You have no access to plb_menu\n");
      }

      RETURN_META(MRES_SUPERCEDE);
   }

   RETURN_META(MRES_HANDLED);
}

// from SDK dlls/cbase.cpp:
// (wd) indicates my comments on the functions
static DLL_FUNCTIONS gFunctionTable =
   {
      NULL, //! pfnGameInit()    Initialize the game (one-time call after loading of game .dll)
      DispatchSpawn, //! pfnSpawn()
      NULL,//DispatchThink,    //! pfnThink()
      NULL,//DispatchUse,    //! pfnUse()
      DispatchTouch,//DispatchTouch,    //! pfnTouch()
      NULL,//DispatchBlocked,   //! pfnBlocked()
      NULL,//DispatchKeyValue,   //! pfnKeyValue()
      NULL,//DispatchSave,    //! pfnSave()
      NULL,//DispatchRestore,   //! pfnRestore()
      NULL,//DispatchObjectCollsionBox, //! pfnSetAbsBox()

      NULL,//SaveWriteFields,   //! pfnSaveWriteFields()
      NULL,//SaveReadFields,    //! pfnSaveReadFields()

      NULL,//SaveGlobalState,   //! pfnSaveGlobalState()
      NULL,//RestoreGlobalState,   //! pfnRestoreGlobalState()
      NULL,//ResetGlobalState,   //! pfnResetGlobalState()

      ClientConnect, //! pfnClientConnect()   (wd) Client has connected
      ClientDisconnect,   //! pfnClientDisconnect()  (wd) Player has left the game
      NULL,//ClientKill,     //! pfnClientKill()    (wd) Player has typed "kill"
      ClientPutInServer,   //! pfnClientPutInServer()  (wd) Client is entering the game
      ClientCommand, //! pfnClientCommand()   (wd) Player has sent a command (typed, or from a bind)
      ClientUserInfoChanged, //! pfnClientUserInfoChanged() (wd) Client has updated their setinfo structure
      NULL,//ServerActivate,    //! pfnServerActivate()   (wd) Server is starting a new map
      NULL,//ServerDeactivate,   //! pfnServerDeactivate()  (wd) Server is leaving the map (shutdown, or changelevel); SDK2

      NULL,//PlayerPreThink,    //! pfnPlayerPreThink()
      NULL,//PlayerPostThink,   //! pfnPlayerPostThink()

      StartFrame, //! pfnStartFrame()
      NULL,//ParmsNewLevel,    //! pfnParmsNewLevel()
      NULL,//ParmsChangeLevel,   //! pfnParmsChangeLevel()

      NULL,//GetGameDescription,   //! pfnGetGameDescription()  Returns string describing current .dll.  E.g. "TeamFotrress 2", "Half-Life"
      NULL,//PlayerCustomization,  //! pfnPlayerCustomization() Notifies .dll of new customization for player.

      NULL,//SpectatorConnect,   //! pfnSpectatorConnect()  Called when spectator joins server
      NULL,//SpectatorDisconnect,  //! pfnSpectatorDisconnect() Called when spectator leaves the server
      NULL,//SpectatorThink,    //! pfnSpectatorThink()   Called when spectator sends a command packet (usercmd_t)

      NULL,//Sys_Error,     //! pfnSys_Error()    Notify game .dll that engine is going to shut down.  Allows mod authors to set a breakpoint.  SDK2

      NULL,//PM_Move,     //! pfnPM_Move()    (wd) SDK2
      NULL,//PM_Init,     //! pfnPM_Init()    Server version of player movement initialization; (wd) SDK2
      NULL,//PM_FindTextureType,   //! pfnPM_FindTextureType()  (wd) SDK2

      NULL,//SetupVisibility,   //! pfnSetupVisibility()  Set up PVS and PAS for networking for this client; (wd) SDK2
      NULL,//UpdateClientData,   //! pfnUpdateClientData()  Set up data sent only to specific client; (wd) SDK2
      NULL,//AddToFullPack,    //! pfnAddToFullPack()   (wd) SDK2
      NULL,//CreateBaseline,    //! pfnCreateBaseline()   Tweak entity baseline for network encoding, allows setup of player baselines, too.; (wd) SDK2
      NULL,//RegisterEncoders,   //! pfnRegisterEncoders()  Callbacks for network encoding; (wd) SDK2
      NULL,//GetWeaponData,    //! pfnGetWeaponData()   (wd) SDK2
      NULL,//CmdStart,     //! pfnCmdStart()    (wd) SDK2
      NULL,//CmdEnd,      //! pfnCmdEnd()     (wd) SDK2
      NULL,//ConnectionlessPacket,  //! pfnConnectionlessPacket() (wd) SDK2
      NULL,//GetHullBounds,    //! pfnGetHullBounds()   (wd) SDK2
      NULL,//CreateInstancedBaselines, //! pfnCreateInstancedBaselines() (wd) SDK2
      NULL,//InconsistentFile,   //! pfnInconsistentFile()  (wd) SDK2
      NULL,//AllowLagCompensation,  //! pfnAllowLagCompensation() (wd) SDK2
   };

// It's not clear what the difference is between GetAPI and GetAPI2; they
// both appear to return the exact same function table.
//
// Only one of them appears to be ever called, though.  If the DLL provides
// GetAPI2, the engine/hlds will call that, and will not call GetAPI.  If
// the engine couldn't find GetAPI2 in the DLL, it appears to fall back to
// GetAPI.
//
// So, GetAPI2 appears to replace GetAPI, and appears to have been added
// with SDK 2.0.  My best guess is that, with the new SDK, interface
// version checking became important, and without the int ptr used in
// GetAPI2, the engine can't find out the version of the DLL via GetAPI.
//
// It's unclear whether a DLL coded under SDK2 needs to provide the older
// GetAPI or not..

C_DLLEXPORT int GetEntityAPI2( DLL_FUNCTIONS *pFunctionTable, int *interfaceVersion )
{
   LOG_DEVELOPER(PLID, "called: GetEntityAPI2; version=%d", *interfaceVersion);
   if(!pFunctionTable)
   {
      LOG_ERROR(PLID, "GetEntityAPI2 called with null pFunctionTable");
      return(FALSE);
   }
   else if(*interfaceVersion != INTERFACE_VERSION)
   {
      LOG_ERROR(PLID, "GetEntityAPI2 version mismatch; requested=%d ours=%d", *interfaceVersion, INTERFACE_VERSION);
      //! Tell engine what version we had, so it can figure out who is out of date.
      *interfaceVersion = INTERFACE_VERSION;
      return(FALSE);
   }
   memcpy( pFunctionTable, &gFunctionTable, sizeof( DLL_FUNCTIONS ) );
   return(TRUE);
}


// I could find _no_ documentation or examples for the intended use of
// NEW_DLL_FUNCTIONS.  I wouldn't have even _known_ about the
// GetNewDLLFunctions() function except for the reference in Adminmod..  It
// appears to be new with SDK 2.0.
//
// Obviously, it seems to provide additional functions to the engine, but
// it's unclear why a new table and interface were added, rather than
// appending new functions to the GetAPI table/interface.
//
// Interestingly, it appears to be called by the engine _before_ GetAPI.

static NEW_DLL_FUNCTIONS gNewFunctionTable =
   {
      NULL,  //! pfnOnFreeEntPrivateData() Called right before the object's memory is freed.  Calls its destructor.
      NULL,  //! pfnGameShutdown()
      NULL,  //! pfnShouldCollide()
   };

C_DLLEXPORT int GetNewDLLFunctions( NEW_DLL_FUNCTIONS *pNewFunctionTable, int *interfaceVersion )
{
   LOG_DEVELOPER(PLID, "called: GetNewDLLFunctions; version=%d", *interfaceVersion);
   if(!pNewFunctionTable)
   {
      LOG_ERROR(PLID, "GetNewDLLFunctions called with null pNewFunctionTable");
      return(FALSE);
   }
   else if(*interfaceVersion != NEW_DLL_FUNCTIONS_VERSION)
   {
      LOG_ERROR(PLID, "GetNewDLLFunctions version mismatch; requested=%d ours=%d", *interfaceVersion, NEW_DLL_FUNCTIONS_VERSION);
      //! Tell engine what version we had, so it can figure out who is out of date.
      *interfaceVersion = NEW_DLL_FUNCTIONS_VERSION;
      return(FALSE);
   }
   memcpy(pNewFunctionTable, &gNewFunctionTable, sizeof(NEW_DLL_FUNCTIONS));
   return TRUE;
}

// Must provide at least one of these..
META_FUNCTIONS gMetaFunctionTable = {
                                       NULL, // pfnGetEntityAPI    HL SDK; called before game DLL
                                       NULL, // pfnGetEntityAPI_Post   META; called after game DLL
                                       GetEntityAPI2, // pfnGetEntityAPI2    HL SDK2; called before game DLL
                                       NULL,//GetEntityAPI2_Post,   // pfnGetEntityAPI2_Post  META; called after game DLL
                                       NULL, // pfnGetNewDLLFunctions  HL SDK2; called before game DLL
                                       NULL,// GetNewDLLFunctions_Post, // pfnGetNewDLLFunctions_Post META; called after game DLL
                                       GetEngineFunctions, // pfnGetEngineFunctions  META; called before HL engine
                                       NULL, // pfnGetEngineFunctions_Post META; called after HL engine
                                    };
